"""Loading example datasets."""
from os.path import dirname, join
import time
import datetime
import io
import numpy as np
import pandas as pd


def load_daily(long: bool = True):
    """Daily sample data.

    wiki = [
            "Germany", "Thanksgiving", 'all', 'Microsoft',
            "Procter_%26_Gamble", "YouTube", "United_States", "Elizabeth_II",
            "William_Shakespeare", "Cleopatra", "George_Washington",
            "Chinese_New_Year", "Standard_deviation", "Christmas",
            "List_of_highest-grossing_films",
            "List_of_countries_that_have_gained_independence_from_the_United_Kingdom",
            "Periodic_table"
    ]

    Sources: Wikimedia Foundation

    Args:
        long (bool): if True, return data in long format. Otherwise return wide
    """
    module_path = dirname(__file__)
    data_file_name = join(module_path, 'data', 'holidays.zip')

    df_wide = pd.read_csv(data_file_name, index_col=0, parse_dates=True)
    if not long:
        return df_wide
    else:
        df_wide.index.name = 'datetime'
        df_long = df_wide.reset_index(drop=False).melt(
            id_vars=['datetime'], var_name='series_id', value_name='value'
        )
        return df_long


def load_fred_monthly():
    """
    Federal Reserve of St. Louis.
    from autots.datasets.fred import get_fred_data
    SeriesNameDict = {'GS10':'10-Year Treasury Constant Maturity Rate',
                              'MCOILWTICO':'Crude Oil West Texas Intermediate Cushing Oklahoma',
                              'CSUSHPISA': ' U.S. National Home Price Index',
                              'EXUSEU': 'US Euro Foreign Exchange Rate',
                              'EXCHUS': 'China US Foreign Exchange Rate',
                              'EXCAUS' : 'Canadian to US Dollar Exchange Rate Daily',
                              'EMVOVERALLEMV': 'Equity Market Volatility Tracker Overall',  # this is a more irregular series
                              'T10YIEM' : '10 Year Breakeven Inflation Rate',
                              'USEPUINDXM': 'Economic Policy Uncertainty Index for United States' # also very irregular
                              }
    monthly_data = get_fred_data(fredkey = 'XXXXXXXXX', SeriesNameDict = SeriesNameDict)
    """
    module_path = dirname(__file__)
    data_file_name = join(module_path, 'data', 'fred_monthly.zip')

    df_long = pd.read_csv(data_file_name, compression='zip')
    df_long['datetime'] = pd.to_datetime(df_long['datetime'])

    return df_long


def load_monthly(long: bool = True):
    """Federal Reserve of St. Louis monthly economic indicators."""
    if long:
        return load_fred_monthly()
    else:
        from autots.tools.shaping import long_to_wide

        df_long = load_fred_monthly()
        df_wide = long_to_wide(
            df_long,
            date_col='datetime',
            value_col='value',
            id_col='series_id',
            aggfunc='first',
        )
        return df_wide


def load_fred_yearly():
    """
    Federal Reserve of St. Louis.
    from autots.datasets.fred import get_fred_data
    SSeriesNameDict = {'GDPA':"Gross Domestic Product",
                  'ACOILWTICO':'Crude Oil West Texas Intermediate Cushing Oklahoma',
                  'AEXUSEU': 'US Euro Foreign Exchange Rate',
                  'AEXCHUS': 'China US Foreign Exchange Rate',
                  'AEXCAUS' : 'Canadian to US Dollar Exchange Rate Daily',
                  'MEHOINUSA672N': 'Real Median US Household Income',
                  'CPALTT01USA661S': 'Consumer Price Index All Items',
                  'FYFSD': 'Federal Surplus or Deficit',
                  'DDDM01USA156NWDB': 'Stock Market Capitalization to US GDP',
                  'LEU0252881600A': 'Median Weekly Earnings for Salary Workers',
                  'LFWA64TTUSA647N': 'US Working Age Population',
                  'IRLTLT01USA156N' : 'Long Term Government Bond Yields'
                  }
    monthly_data = get_fred_data(fredkey = 'XXXXXXXXX', SeriesNameDict = SeriesNameDict)
    """
    module_path = dirname(__file__)
    data_file_name = join(module_path, 'data', 'fred_yearly.zip')

    df_long = pd.read_csv(data_file_name)
    df_long['datetime'] = pd.to_datetime(df_long['datetime'])

    return df_long


def load_yearly(long: bool = True):
    """Federal Reserve of St. Louis annual economic indicators."""
    if long:
        return load_fred_yearly()
    else:
        from autots.tools.shaping import long_to_wide

        df_long = load_fred_yearly()
        df_wide = long_to_wide(
            df_long,
            date_col='datetime',
            value_col='value',
            id_col='series_id',
            aggfunc='first',
        )
        return df_wide


def load_traffic_hourly(long: bool = True):
    """
    From the MN DOT via the UCI data repository.
    Yes, Minnesota is the best state of the Union.
    """
    module_path = dirname(__file__)
    data_file_name = join(module_path, 'data', 'traffic_hourly.zip')

    df_wide = pd.read_csv(
        data_file_name, index_col=0, parse_dates=True, compression='zip'
    )
    if not long:
        return df_wide
    else:
        df_long = df_wide.reset_index(drop=False).melt(
            id_vars=['datetime'], var_name='series_id', value_name='value'
        )
        return df_long


def load_hourly(long: bool = True):
    """Traffic data from the MN DOT via the UCI data repository."""
    return load_traffic_hourly(long=long)


def load_eia_weekly():
    """Weekly petroleum industry data from the EIA."""
    module_path = dirname(__file__)
    data_file_name = join(module_path, 'data', 'eia_weekly.zip')

    df_long = pd.read_csv(data_file_name, compression='zip')
    df_long['datetime'] = pd.to_datetime(df_long['datetime'])
    return df_long


def load_weekly(long: bool = True):
    """Weekly petroleum industry data from the EIA."""
    if long:
        return load_eia_weekly()
    else:
        from autots.tools.shaping import long_to_wide

        df_long = load_eia_weekly()
        df_wide = long_to_wide(
            df_long,
            date_col='datetime',
            value_col='value',
            id_col='series_id',
            aggfunc='first',
        )
        return df_wide


def load_weekdays(long: bool = False, categorical: bool = True, periods: int = 180):
    """Test edge cases by creating a Series with values as day of week.

    Args:
        long (bool):
            if True, return a df with columns "value" and "datetime"
            if False, return a Series with dt index
        categorical (bool): if True, return str/object, else return int
        periods (int): number of periods, ie length of data to generate
    """
    idx = pd.date_range(end=pd.Timestamp.today(), periods=periods, freq="D")
    df_wide = pd.Series(idx.weekday, index=idx, name="value")
    df_wide.index.name = "datetime"
    if categorical:
        df_wide = df_wide.replace(
            {
                0: "Mon",
                1: "Tues",
                2: "Wed",
                3: "Thor's",
                4: "Fri",
                5: "Sat",
                6: "Sun",
                7: "Mon",
            }
        )
    if long:
        return df_wide.reset_index()
    else:
        return df_wide


def load_live_daily(
    long: bool = False,
    observation_start: str = None,
    observation_end: str = None,
    fred_key: str = None,
    fred_series=["DGS10", "T5YIE", "SP500", "DCOILWTICO", "DEXUSEU", "WPU0911"],
    tickers: list = ["MSFT"],
    trends_list: list = ["forecasting", "cycling", "microsoft"],
    trends_geo: str = "US",
    weather_data_types: list = ["AWND", "WSF2", "TAVG"],
    weather_stations: list = ["USW00094846", "USW00014925"],
    weather_years: int = 10,
    london_air_stations: list = ['CT3', 'SK8'],
    london_air_species: str = "PM25",
    london_air_days: int = 180,
    earthquake_days: int = 180,
    earthquake_min_magnitude: int = 5,
    gsa_key: str = None,  # https://open.gsa.gov/api/dap/
    gov_domain_list=['nasa.gov'],
    gov_domain_limit: int = 600,
    wikipedia_pages: list = ['Microsoft_Office', "List_of_highest-grossing_films"],
    wiki_language: str = "en",
    weather_event_types=["%28Z%29+Winter+Weather", "%28Z%29+Winter+Storm"],
    timeout: float = 300.05,
    sleep_seconds: int = 2,
):
    """Generates a dataframe of data up to the present day. Requires active internet connection.
    Try to be respectful of these free data sources by not calling too much too heavily.
    Pass None instead of specification lists to exclude a data source.

    Args:
        long (bool): whether to return in long format or wide
        observation_start (str): %Y-%m-%d earliest day to retrieve, passed to Fred.get_series and yfinance.history
            note that apis with more restricts have other default lengths below which ignore this
        observation_end (str):  %Y-%m-%d most recent day to retrieve
        fred_key (str): https://fred.stlouisfed.org/docs/api/api_key.html
        fred_series (list): list of FRED series IDs. This requires fredapi package
        tickers (list): list of stock tickers, requires yfinance pypi package
        trends_list (list): list of search keywords, requires pytrends pypi package. None to skip.
        weather_data_types (list): from NCEI NOAA api data types, GHCN Daily Weather Elements
            PRCP, SNOW, TMAX, TMIN, TAVG, AWND, WSF1, WSF2, WSF5, WSFG
        weather_stations (list): from NCEI NOAA api station ids. Pass empty list to skip.
        london_air_stations (list): londonair.org.uk source station IDs. Pass empty list to skip.
        london_species (str): what measurement to pull from London Air. Not all stations have all metrics.
        earthquake_min_magnitude (int): smallest earthquake magnitude to pull from earthquake.usgs.gov. Set None to skip this.
        gsa_key (str): api key from https://open.gsa.gov/api/dap/
        gov_domain_list (list): dist of government run domains to get traffic data for. Can be very slow, so fewer is better.
            some examples: ['usps.com', 'ncbi.nlm.nih.gov', 'cdc.gov', 'weather.gov', 'irs.gov', "usajobs.gov", "studentaid.gov", 'nasa.gov', "uk.usembassy.gov", "tsunami.gov"]
        gov_domain_limit (int): max number of records. Smaller will be faster. Max is currently 10000.
        wikipedia_pages (list): list of Wikipedia pages, html encoded if needed (underscore for space)
        weather_event_types (list): list of html encoded severe weather event types https://www1.ncdc.noaa.gov/pub/data/swdi/stormevents/csvfiles/Storm-Data-Export-Format.pdf
        timeout (float): used by some queries
        sleep_seconds (int): increasing this may reduce probability of server download failures
    """
    assert sleep_seconds >= 0.5, "sleep_seconds must be >=0.5"

    dataset_lists = []
    if observation_end is None:
        current_date = datetime.datetime.utcnow()
    else:
        current_date = observation_end
    if observation_start is None:
        # should take from observation_end but that's expected as a string
        observation_start = datetime.datetime.utcnow() - datetime.timedelta(
            days=365 * 6
        )
        observation_start = observation_start.strftime("%Y-%m-%d")
    try:
        import requests

        s = requests.Session()
    except Exception as e:
        print(f"requests Session creation failed {repr(e)}")

    try:
        if fred_key is not None and fred_series is not None:
            from fredapi import Fred  # noqa
            from autots.datasets.fred import get_fred_data

            fred_df = get_fred_data(
                fred_key,
                fred_series,
                long=False,
                observation_start=observation_start,
                sleep_seconds=sleep_seconds,
            )
            fred_df.index = fred_df.index.tz_localize(None)
            dataset_lists.append(fred_df)
    except ModuleNotFoundError:
        print("pip install fredapi (and you'll also need an api key)")
    except Exception as e:
        print(f"FRED data failed: {repr(e)}")

    if tickers is not None:
        for ticker in tickers:
            try:
                import yfinance as yf

                msft = yf.Ticker(ticker)
                # get historical market data
                msft_hist = msft.history(start=observation_start)
                msft_hist = msft_hist.rename(
                    columns=lambda x: x.lower().replace(" ", "_")
                )
                msft_hist = msft_hist.rename(columns=lambda x: ticker.lower() + "_" + x)
                try:
                    msft_hist.index = msft_hist.index.tz_localize(None)
                except Exception:
                    pass
                dataset_lists.append(msft_hist)
                time.sleep(sleep_seconds)
            except ModuleNotFoundError:
                print("You need to: pip install yfinance")
            except Exception as e:
                print(f"yfinance data failed: {repr(e)}")

    str_end_time = current_date.strftime("%Y-%m-%d")
    start_date = (current_date - datetime.timedelta(days=360 * weather_years)).strftime(
        "%Y-%m-%d"
    )
    if weather_stations is not None:
        for wstation in weather_stations:
            try:
                wbase = "https://www.ncei.noaa.gov/access/services/data/v1/?dataset=daily-summaries"
                wargs = f"&dataTypes={','.join(weather_data_types)}&stations={wstation}"
                wargs = (
                    wargs
                    + f"&startDate={start_date}&endDate={str_end_time}&boundingBox=90,-180,-90,180&units=standard&format=csv"
                )
                wdf = pd.read_csv(
                    io.StringIO(s.get(wbase + wargs, timeout=timeout).text)
                )
                wdf['DATE'] = pd.to_datetime(wdf['DATE'])
                wdf = wdf.set_index('DATE').drop(columns=['STATION'])
                wdf.rename(columns=lambda x: wstation + "_" + x, inplace=True)
                dataset_lists.append(wdf)
                time.sleep(sleep_seconds)
            except Exception as e:
                print(f"weather data failed: {repr(e)}")

    str_end_time = current_date.strftime("%d-%b-%Y")
    start_date = (current_date - datetime.timedelta(days=london_air_days)).strftime(
        "%d-%b-%Y"
    )
    if london_air_stations is not None:
        for asite in london_air_stations:
            try:
                # abase = "http://api.erg.ic.ac.uk/AirQuality/Data/Site/Wide/"
                # aargs = "SiteCode=CT8/StartDate=2021-07-01/EndDate=2021-07-30/csv"
                abase = 'https://www.londonair.org.uk/london/asp/downloadsite.asp'
                aargs = f"?site={asite}&species1={london_air_species}m&species2=&species3=&species4=&species5=&species6=&start={start_date}&end={str_end_time}&res=6&period=daily&units=ugm3"
                data = s.get(abase + aargs, timeout=timeout).content
                adf = pd.read_csv(io.StringIO(data.decode('utf-8')))
                acol = adf['Site'].iloc[0] + "_" + adf['Species'].iloc[0]
                adf['Datetime'] = pd.to_datetime(adf['ReadingDateTime'], dayfirst=True)
                adf[acol] = adf['Value']
                dataset_lists.append(adf[['Datetime', acol]].set_index("Datetime"))
                time.sleep(sleep_seconds)
                # "/Data/Traffic/Site/SiteCode={SiteCode}/StartDate={StartDate}/EndDate={EndDate}/Json"
            except Exception as e:
                print(f"London Air data failed: {repr(e)}")

    if earthquake_min_magnitude is not None:
        try:
            str_end_time = current_date.strftime("%Y-%m-%d")
            start_date = (
                current_date - datetime.timedelta(days=earthquake_days)
            ).strftime("%Y-%m-%d")
            # is limited to ~1000 rows of data, ie individual earthquakes
            ebase = "https://earthquake.usgs.gov/fdsnws/event/1/query?"
            eargs = f"format=csv&starttime={start_date}&endtime={str_end_time}&minmagnitude={earthquake_min_magnitude}"
            eq = pd.read_csv(ebase + eargs)
            eq["time"] = pd.to_datetime(eq["time"])
            eq["time"] = eq["time"].dt.tz_localize(None)
            eq.set_index("time", inplace=True)
            global_earthquakes = eq.resample("1D").agg(
                {"mag": "mean", "depth": "count"}
            )
            global_earthquakes["mag"] = global_earthquakes["mag"].fillna(
                earthquake_min_magnitude
            )
            global_earthquakes = global_earthquakes.rename(
                columns={
                    "mag": "largest_magnitude_earthquake",
                    "depth": "count_large_earthquakes",
                }
            )
            dataset_lists.append(global_earthquakes)
        except Exception as e:
            print(f"earthquake data failed: {repr(e)}")

    if gov_domain_list is not None:
        try:
            # print because this one is slow, and point people at that fact
            if gsa_key is None:
                gsa_key = "DEMO_KEY2"
            # only run 1 if demo_key1
            if "DEMO_KEY" in gsa_key:
                gov_domain_list = gov_domain_list[0:1]
            for domain in gov_domain_list:
                report = "domain"  # site, domain, download, second-level-domain
                url = f"https://api.gsa.gov/analytics/dap/v1.1/domain/{domain}/reports/{report}/data?api_key={gsa_key}&limit={gov_domain_limit}&after={observation_start}"
                data = s.get(url, timeout=timeout)
                gdf = pd.read_json(data.text, orient="records")
                gdf['date'] = pd.to_datetime(gdf['date'])
                # essentially duplicates brought by agency and null agency
                gresult = gdf.groupby('date')['visits'].first()
                gresult.name = domain
                dataset_lists.append(gresult.to_frame())
                time.sleep(sleep_seconds)
        except Exception as e:
            print(f"analytics.gov data failed with {repr(e)}")

    if wikipedia_pages is not None:
        str_start = pd.to_datetime(observation_start).strftime("%Y%m%d00")
        str_end = current_date.strftime("%Y%m%d00")
        headers = {
            'User-Agent': 'AutoTS load_live_daily',
        }
        for page in wikipedia_pages:
            try:
                if page == "all":
                    url = f"https://wikimedia.org/api/rest_v1/metrics/pageviews/aggregate/all-projects/all-access/all-agents/daily/{str_start}/{str_end}?maxlag=5"
                else:
                    url = f"https://wikimedia.org/api/rest_v1/metrics/pageviews/per-article/{wiki_language}.wikipedia/all-access/all-agents/{page}/daily/{str_start}/{str_end}?maxlag=5"
                data = s.get(url, timeout=timeout, headers=headers)
                data_js = data.json()
                if "items" not in data_js.keys():
                    print(data_js)
                gdf = pd.DataFrame(data_js['items'])
                gdf['date'] = pd.to_datetime(gdf['timestamp'], format="%Y%m%d00")
                gresult = gdf.set_index('date')['views'].fillna(0)
                gresult.name = "wiki_" + str(page)[0:80]
                dataset_lists.append(gresult.to_frame())
                time.sleep(sleep_seconds)
            except Exception as e:
                print(f"Wikipedia api failed with error {repr(e)}")
                time.sleep(10)

    if weather_event_types is not None:
        try:
            for event_type in weather_event_types:
                # appears to have a fixed max of 500 records
                url = f"https://www.ncdc.noaa.gov/stormevents/csv?eventType={event_type}&beginDate_mm=01&beginDate_dd=01&beginDate_yyyy=2000&endDate_mm=09&endDate_dd=30&endDate_yyyy=9999&hailfilter=0.00&tornfilter=2&windfilter=000&sort=DN&statefips=-999%2CALL"
                csv_in = io.StringIO(s.get(url, timeout=timeout).text)
                try:
                    # new in 1.3.0 of pandas
                    df = pd.read_csv(csv_in, low_memory=False, on_bad_lines='skip')
                except Exception:
                    df = pd.read_csv(csv_in, low_memory=False, error_bad_lines=False)
                df['BEGIN_DATE'] = pd.to_datetime(df['BEGIN_DATE'])
                df['END_DATE'] = pd.to_datetime(df['END_DATE'])
                df['day'] = df.apply(
                    lambda row: pd.date_range(
                        row["BEGIN_DATE"], row['END_DATE'], freq='D'
                    ),
                    axis=1,
                )
                df = df.explode('day')
                swresult = df.groupby(["day"])["EVENT_ID"].count()
                swresult.name = "_".join(event_type.split("+")[1:]) + "_Events"
                dataset_lists.append(swresult.to_frame())
                time.sleep(sleep_seconds)
        except Exception as e:
            print(f"Severe Weather data failed with {repr(e)}")

    if trends_list is not None:
        try:
            from pytrends.request import TrendReq

            pytrends = TrendReq(hl="en-US", tz=360)
            pytrends.build_payload(trends_list, geo=trends_geo)
            # pytrends.build_payload(trends_list, timeframe="all")  # 'today 12-m'
            gtrends = pytrends.interest_over_time()
            gtrends.index = gtrends.index.tz_localize(None)
            gtrends.drop(columns="isPartial", inplace=True, errors="ignore")
            dataset_lists.append(gtrends)
        except ImportError:
            print("You need to: pip install pytrends")
        except Exception as e:
            print(f"pytrends data failed: {repr(e)}")

    if len(dataset_lists) < 1:
        raise ValueError("No data successfully downloaded!")
    elif len(dataset_lists) == 1:
        df = dataset_lists[0]
    else:
        from functools import reduce

        df = reduce(
            lambda x, y: pd.merge(x, y, left_index=True, right_index=True, how="outer"),
            dataset_lists,
        )
    print(f"{df.shape[1]} series downloaded.")
    s.close()

    if not long:
        return df
    else:
        df_long = df.reset_index(drop=False).melt(
            id_vars=['datetime'], var_name='series_id', value_name='value'
        )
        return df_long


def load_zeroes(long=False, shape=None, start_date: str = "2021-01-01"):
    """Create a dataset of just zeroes for testing edge case."""
    if shape is None:
        shape = (200, 5)
    df_wide = pd.DataFrame(
        np.zeros(shape), index=pd.date_range(start_date, periods=shape[0], freq="D")
    )
    if not long:
        return df_wide
    else:
        df_wide.index.name = "datetime"
        df_long = df_wide.reset_index(drop=False).melt(
            id_vars=['datetime'], var_name='series_id', value_name='value'
        )
        return df_long


def load_linear(
    long=False,
    shape=None,
    start_date: str = "2021-01-01",
    introduce_nan: float = None,
    introduce_random: float = None,
    random_seed: int = 123,
):
    """Create a dataset of just zeroes for testing edge case.

    Args:
        long (bool): whether to make long or wide
        shape (tuple): shape of output dataframe
        start_date (str): first date of index
        introduce_nan (float): percent of rows to make null. 0.2 = 20%
        introduce_random (float): shape of gamma distribution
        random_seed (int): seed for random
    """
    if shape is None:
        shape = (500, 5)
    idx = pd.date_range(start_date, periods=shape[0], freq="D")
    df_wide = pd.DataFrame(np.ones(shape), index=idx)
    df_wide = (df_wide * list(range(0, shape[1]))).cumsum()
    if introduce_nan is not None:
        df_wide = df_wide.sample(
            frac=(1 - introduce_nan), random_state=random_seed
        ).reindex(idx)
    if introduce_random is not None:
        df_wide = df_wide + np.random.default_rng(random_seed).gamma(
            introduce_random, size=shape
        )
    if not long:
        return df_wide
    else:
        df_wide.index.name = "datetime"
        df_long = df_wide.reset_index(drop=False).melt(
            id_vars=['datetime'], var_name='series_id', value_name='value'
        )
        return df_long


def load_sine(
    long=False,
    shape=None,
    start_date: str = "2021-01-01",
    introduce_random: float = None,
    random_seed: int = 123,
):
    """Create a dataset of just zeroes for testing edge case."""
    if shape is None:
        shape = (500, 5)
    df_wide = pd.DataFrame(
        np.ones(shape),
        index=pd.date_range(start_date, periods=shape[0], freq="D"),
        columns=range(shape[1]),
    )
    X = pd.to_numeric(df_wide.index, errors='coerce', downcast='integer').values

    def sin_func(a, X):
        return a * np.sin(a * X) + a

    for column in df_wide.columns:
        df_wide[column] = sin_func(column, X)
    if introduce_random is not None:
        df_wide = (
            df_wide
            + np.random.default_rng(random_seed).gamma(introduce_random, size=shape)
        ).clip(lower=0.1)
    if not long:
        return df_wide
    else:
        df_wide.index.name = "datetime"
        df_long = df_wide.reset_index(drop=False).melt(
            id_vars=['datetime'], var_name='series_id', value_name='value'
        )
        return df_long


def load_artificial(long=False, date_start=None, date_end=None):
    """Load artifically generated series from random distributions.

    Args:
        long (bool): if True long style data, if False, wide style data
        date_start: str or datetime.datetime of start date
        date_end: str or datetime.datetime of end date
    """
    import scipy.signal
    from scipy.ndimage import maximum_filter1d

    if date_end is None:
        date_end = datetime.datetime.now().date()
    if isinstance(date_end, datetime.datetime):
        date_end = date_end.date()
    if date_start is None:
        if isinstance(date_end, datetime.date):
            date_start = date_end - datetime.timedelta(days=720)
        else:
            date_start = datetime.datetime.now().date() - datetime.timedelta(days=720)
    if isinstance(date_start, datetime.datetime):
        date_start = date_start.date()
    dates = pd.date_range(date_start, date_end)
    size = dates.size
    rng = np.random.default_rng()

    df_wide = pd.DataFrame(
        {
            'white_noise': rng.normal(0, 1, size),
            "white_noise_trend": rng.normal(0, 1, size) + np.arange(size) * 0.01,
            "random_walk": np.random.choice(a=[-0.8, 0, 0.8], size=size).cumsum() * 0.8,
            "arima007_trend": np.convolve(
                np.random.choice(a=[-0.4, 0, 0.4], size=size + 6),
                np.ones(7, dtype=int),
                'valid',
            )
            + np.arange(size) * 0.01,
            "arima017": np.convolve(
                np.random.choice(a=[-0.4, 0, 0.4], size=size + 6),
                np.ones(7, dtype=int),
                'valid',
            ).cumsum()
            / 12,
            "arima200_gamma": scipy.signal.lfilter(
                [1], [1.0, -0.75, 0.25], 1 * rng.gamma(1, size=size), axis=0
            ),  # ma order is first, then ar order
            "arima220_outliers": np.where(
                rng.poisson(20, size) >= 30,
                rng.gamma(5, size=size) + 10,
                scipy.signal.lfilter(
                    [1.0, 0.65, 0.25],
                    [1.0, -0.85, 0.25],
                    1 * rng.normal(2, size=size),
                    axis=0,
                )
                / 2,
            ),
            "linear": np.arange(size) * 0.025,
            "sine_wave": np.sin(np.arange(size)),
            "sine_seasonality_monthweek": (
                (np.sin((np.pi / 7) * np.arange(size)) * 0.25 + 0.25)
                + (np.sin((np.pi / 28) * np.arange(size)) * 1 + 1)
                + rng.normal(0, 0.15, size)
            ),
            "wavelet_ricker": np.tile(
                scipy.signal.ricker(33, 1), int(np.ceil(size / 33))
            )[:size],
            "wavelet_morlet": np.real(
                np.tile(scipy.signal.morlet2(100, 6.0, 6.0), int(np.ceil(size / 100)))[
                    :size
                ]
                * 10
            ),
            "lumpy": np.stack(
                [
                    rng.gamma(1, size=size),
                    rng.gamma(1, size=size),
                    rng.gamma(1, size=size),
                    rng.gamma(4, size=size),
                    rng.gamma(6, size=size),
                    rng.gamma(7, size=size),
                    rng.gamma(5, size=size),
                ],
                axis=0,
            ).T.ravel()[:size]
            + (np.sin((np.pi / 182) * np.arange(size)) * 0.75 + 1),
            "intermittent_random": rng.poisson(0.3, size=size),
            "intermittent_weekly": np.stack(
                [
                    np.random.choice(a=[0, 1], p=[0.98, 0.02], size=size),
                    np.random.choice(a=[0, 1], p=[0.96, 0.04], size=size),
                    np.random.choice(a=[0, 1], p=[0.94, 0.06], size=size),
                    np.random.choice(a=[0, 1], p=[0.94, 0.06], size=size),
                    np.random.choice(a=[0, 2, 1], p=[0.8, 0.1, 0.1], size=size),
                    np.random.choice(
                        a=[0, 3, 2, 1], p=[0.5, 0.05, 0.1, 0.35], size=size
                    ),
                    np.random.choice(
                        a=[0, 3, 2, 1], p=[0.25, 0.2, 0.3, 0.25], size=size
                    ),
                ],
                axis=0,
            ).T.ravel()[:size],
            "out_of_stock": np.where(
                -maximum_filter1d(-rng.negative_binomial(1, 0.04, size=size), 8) == 0,
                0,
                # moving average of a sine + gamma random
                np.convolve(
                    (
                        (np.sin((np.pi / 182) * np.arange(size + 2)) * 2 + 2)
                        + rng.gamma(1, 0.5, size=size + 2)
                    ),
                    np.ones(3, dtype=int),
                    'valid',
                )
                / 2,
            ),
            "cubic_root": np.cbrt(np.arange(-int(size / 2), size - int(size / 2))),
            "logistic_growth": np.log(np.arange(2, size + 2)),
            "recent_spike": np.where(
                np.arange(size) < (9 * size) / 10,
                np.arange(size) * 0.01,
                abs(np.arange(size) - (9 * size) / 10) ** 1.01 + (9 * size) / 10 * 0.01,
            )
            + rng.normal(0, 0.05, size),
            "recent_plateau": np.where(
                np.arange(size) < (8.5 * size) / 10,
                np.arange(size) * 0.01,
                (8.5 * size * 0.01) / 10,
            )
            + rng.normal(0, 0.05, size),
            "old_to_new": np.where(
                np.arange(size) < (4 * size) / 5,
                np.real(
                    np.tile(
                        scipy.signal.morlet2(50, 6.0, 6.0), int(np.ceil(size / 50))
                    )[:size]
                    * 10
                ),
                np.real(
                    np.tile(
                        scipy.signal.morlet2(50, 6.0, 0.0), int(np.ceil(size / 50))
                    )[:size]
                    * 10
                ),
            )
            + (np.sin((np.pi / 182) * np.arange(size)) * 1 + 1),
        },
        index=dates,
    )

    if not long:
        return df_wide
    else:
        df_wide.index.name = "datetime"
        df_long = df_wide.reset_index(drop=False).melt(
            id_vars=['datetime'], var_name='series_id', value_name='value'
        )
        return df_long



"""Neural Nets."""
import pandas as pd
from autots.tools.shaping import wide_to_3d

try:
    import tensorflow as tf
except Exception:  # except ImportError
    _has_tf = False
else:
    _has_tf = True
"""
X, Y = window_maker(df, forecast_length = 6, shuffle = False,
                    input_dim = 'univariate', window_size = 10,
                    output_dim = '1step')
"""

if _has_tf:

    class ResidualWrapper(tf.keras.Model):
        """From https://www.tensorflow.org/tutorials/structured_data/time_series"""

        def __init__(self, model):
            super().__init__()
            self.model = model

        def call(self, inputs, *args, **kwargs):
            delta = self.model(inputs, *args, **kwargs)

            # The prediction for each time step is the input
            # from the previous time step plus the delta
            # calculated by the model.
            return inputs + delta


class KerasRNN(object):
    """Wrapper for Tensorflow Keras based RNN.

    Args:
        rnn_type (str): Keras cell type 'GRU' or default 'LSTM'
        kernel_initializer (str): passed to first keras LSTM or GRU layer
        hidden_layer_sizes (tuple): of len 1 or 3 passed to first keras LSTM or GRU layers
        optimizer (str): Passed to keras model.compile
        loss (str): Passed to keras model.compile
        epochs (int): Passed to keras model.fit
        batch_size (int): Passed to keras model.fit
        verbose (int): 0, 1 or 2. Passed to keras model.fit
        random_seed (int): passed to tf.random.set_seed()
    """

    def __init__(
        self,
        rnn_type: str = 'LSTM',
        kernel_initializer: str = 'lecun_uniform',
        hidden_layer_sizes: tuple = (32, 32, 32),
        optimizer: str = 'adam',
        loss: str = 'huber',
        epochs: int = 50,
        batch_size: int = 32,
        shape=1,
        verbose: int = 1,
        random_seed: int = 2020,
    ):
        self.name = 'KerasRNN'
        verbose = 0 if verbose < 0 else verbose
        verbose = 2 if verbose > 2 else verbose
        self.verbose = verbose
        self.random_seed = random_seed
        self.kernel_initializer = kernel_initializer
        self.epochs = epochs
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.loss = loss
        self.hidden_layer_sizes = hidden_layer_sizes
        self.rnn_type = rnn_type
        self.shape = shape

    def fit(self, X, Y):
        """Train the model on dataframes of X and Y."""
        if not _has_tf:
            raise ImportError(
                "Tensorflow not available, install with pip install tensorflow."
            )
        tf.keras.backend.clear_session()
        tf.random.set_seed(self.random_seed)
        train_X = pd.DataFrame(X).to_numpy()
        # target shape is [samples, timesteps, features]
        if self.shape == 1:
            train_X = train_X.reshape((train_X.shape[0], 1, train_X.shape[1]))
        elif self.shape > 2:
            # needs matching Y shape and predict input
            train_X = wide_to_3d(train_X, self.shape)
        else:
            train_X = train_X.reshape((train_X.shape[0], train_X.shape[1], 1))
        INPUT_SHAPE = (train_X.shape[1], train_X.shape[2])
        if Y.ndim == 1:
            Y = Y.reshape(-1, 1)
        OUTPUT_SHAPE = Y.shape[1]
        if self.rnn_type == "E2D2":
            # crudely borrowed from: https://www.analyticsvidhya.com/blog/2020/10/multivariate-multi-step-time-series-forecasting-using-stacked-lstm-sequence-to-sequence-autoencoder-in-tensorflow-2-0-keras/
            encoder_inputs = tf.keras.layers.Input(shape=INPUT_SHAPE)
            encoder_l1 = tf.keras.layers.LSTM(
                self.hidden_layer_sizes[0], return_sequences=True, return_state=True
            )
            encoder_outputs1 = encoder_l1(encoder_inputs)
            encoder_states1 = encoder_outputs1[1:]
            encoder_l2 = tf.keras.layers.LSTM(
                self.hidden_layer_sizes[0], return_state=True
            )
            encoder_outputs2 = encoder_l2(encoder_outputs1[0])
            encoder_states2 = encoder_outputs2[1:]
            #
            decoder_inputs = tf.keras.layers.RepeatVector(OUTPUT_SHAPE)(
                encoder_outputs2[0]
            )
            layer_2_shape = self.hidden_layer_sizes
            layer_2_size = (
                layer_2_shape[2] if len(layer_2_shape) >= 3 else layer_2_shape[0]
            )
            #
            decoder_l1 = tf.keras.layers.LSTM(layer_2_size, return_sequences=True)(
                decoder_inputs, initial_state=encoder_states1
            )
            decoder_l2 = tf.keras.layers.LSTM(layer_2_size, return_sequences=True)(
                decoder_l1, initial_state=encoder_states2
            )
            decoder_outputs2 = tf.keras.layers.TimeDistributed(
                tf.keras.layers.Dense(OUTPUT_SHAPE)
            )(decoder_l2)
            #
            self.model = tf.keras.models.Model(encoder_inputs, decoder_outputs2)
        if self.rnn_type == "CNN":
            if len(self.hidden_layer_sizes) == 1:
                kernel_size = 10 if INPUT_SHAPE[0] > 10 else INPUT_SHAPE[0]
                self.model = tf.keras.Sequential(
                    [
                        tf.keras.layers.Conv1D(
                            filters=self.hidden_layer_sizes[0],
                            kernel_size=kernel_size,
                            activation='relu',
                            input_shape=INPUT_SHAPE,
                            kernel_initializer=self.kernel_initializer,
                        ),
                        tf.keras.layers.Dense(
                            units=self.hidden_layer_sizes[0], activation='relu'
                        ),
                        tf.keras.layers.Dense(units=OUTPUT_SHAPE),
                    ]
                )
            else:
                # borrowed from https://keras.io/examples/timeseries/timeseries_classification_from_scratch
                input_layer = tf.keras.layers.Input(INPUT_SHAPE)
                layer_shape = self.hidden_layer_sizes
                layer_2_size = (
                    layer_shape[1] if len(layer_shape) >= 2 else layer_shape[0]
                )
                layer_3_size = (
                    layer_shape[2] if len(layer_shape) >= 3 else layer_shape[0]
                )

                conv1 = tf.keras.layers.Conv1D(
                    filters=self.hidden_layer_sizes[0],
                    kernel_size=3,
                    padding="same",
                    kernel_initializer=self.kernel_initializer,
                )(input_layer)
                conv1 = tf.keras.layers.BatchNormalization()(conv1)
                conv1 = tf.keras.layers.ReLU()(conv1)
                conv2 = tf.keras.layers.Conv1D(
                    filters=layer_2_size, kernel_size=3, padding="same"
                )(conv1)
                conv2 = tf.keras.layers.BatchNormalization()(conv2)
                conv2 = tf.keras.layers.ReLU()(conv2)
                conv3 = tf.keras.layers.Conv1D(
                    filters=layer_3_size, kernel_size=3, padding="same"
                )(conv2)
                conv3 = tf.keras.layers.BatchNormalization()(conv3)
                conv3 = tf.keras.layers.ReLU()(conv3)

                gap = tf.keras.layers.GlobalAveragePooling1D()(conv3)
                output_layer = tf.keras.layers.Dense(OUTPUT_SHAPE)(gap)
                self.model = tf.keras.models.Model(
                    inputs=input_layer, outputs=output_layer
                )
        elif len(self.hidden_layer_sizes) == 3:
            if self.rnn_type == 'GRU':
                self.model = tf.keras.models.Sequential(
                    [
                        tf.keras.layers.GRU(
                            self.hidden_layer_sizes[0],
                            kernel_initializer=self.kernel_initializer,
                            input_shape=INPUT_SHAPE,
                            return_sequences=True,
                        ),
                        tf.keras.layers.GRU(
                            self.hidden_layer_sizes[1], return_sequences=True
                        ),
                        tf.keras.layers.Dropout(0.2),
                        tf.keras.layers.GRU(self.hidden_layer_sizes[2]),
                        tf.keras.layers.Dense(OUTPUT_SHAPE),
                    ]
                )
            else:
                self.model = tf.keras.models.Sequential(
                    [
                        tf.keras.layers.LSTM(
                            self.hidden_layer_sizes[0],
                            kernel_initializer=self.kernel_initializer,
                            input_shape=INPUT_SHAPE,
                            return_sequences=True,
                        ),
                        tf.keras.layers.Dropout(0.2),
                        tf.keras.layers.LSTM(
                            self.hidden_layer_sizes[1], return_sequences=True
                        ),
                        tf.keras.layers.Dropout(0.2),
                        tf.keras.layers.LSTM(self.hidden_layer_sizes[2]),
                        tf.keras.layers.Dense(OUTPUT_SHAPE),
                    ]
                )
        if len(self.hidden_layer_sizes) == 1:
            if self.rnn_type == 'GRU':
                self.model = ResidualWrapper(
                    tf.keras.models.Sequential(
                        [
                            tf.keras.layers.GRU(
                                self.hidden_layer_sizes[0],
                                kernel_initializer=self.kernel_initializer,
                                input_shape=INPUT_SHAPE,
                            ),
                            tf.keras.layers.Dense(10, activation='relu'),
                            tf.keras.layers.Dense(OUTPUT_SHAPE),
                        ]
                    )
                )
            else:
                self.model = tf.keras.models.Sequential(
                    [
                        tf.keras.layers.Bidirectional(
                            tf.keras.layers.LSTM(
                                self.hidden_layer_sizes[0],
                                kernel_initializer=self.kernel_initializer,
                                input_shape=INPUT_SHAPE,
                            )
                        ),
                        tf.keras.layers.Dense(32, activation='relu'),
                        tf.keras.layers.Dense(OUTPUT_SHAPE),
                        tf.keras.layers.Lambda(lambda x: x * 100.0),
                    ]
                )

        if str(self.loss).lower() == 'huber':
            loss = tf.keras.losses.Huber()
        else:
            loss = self.loss
        self.model.compile(optimizer=self.optimizer, loss=loss)

        self.model.fit(
            x=train_X,
            y=Y,
            epochs=self.epochs,
            batch_size=self.batch_size,
            verbose=self.verbose,
        )
        return self

    def predict(self, X):
        """Predict on dataframe of X."""
        if self.shape == 1:
            test = pd.DataFrame(X).to_numpy().reshape((X.shape[0], 1, X.shape[1]))
        else:
            test = pd.DataFrame(X).to_numpy().reshape((X.shape[0], X.shape[1], 1))
        return self.model.predict(test)


def transformer_encoder(inputs, head_size, num_heads, ff_dim, dropout=0):
    # Attention and Normalization
    x = tf.keras.layers.MultiHeadAttention(
        key_dim=head_size, num_heads=num_heads, dropout=dropout
    )(inputs, inputs)
    x = tf.keras.layers.Dropout(dropout)(x)
    x = tf.keras.layers.LayerNormalization(epsilon=1e-6)(x)
    res = x + inputs

    # Feed Forward Part
    x = tf.keras.layers.Conv1D(filters=ff_dim, kernel_size=1, activation="relu")(res)
    x = tf.keras.layers.Dropout(dropout)(x)
    x = tf.keras.layers.Conv1D(filters=inputs.shape[-1], kernel_size=1)(x)
    x = tf.keras.layers.LayerNormalization(epsilon=1e-6)(x)
    return x + res


def transformer_build_model(
    input_shape,
    output_shape,
    head_size,
    num_heads,
    ff_dim,
    num_transformer_blocks,
    mlp_units,
    dropout=0,
    mlp_dropout=0,
):
    inputs = tf.keras.Input(shape=input_shape)
    x = inputs
    for _ in range(num_transformer_blocks):
        x = transformer_encoder(x, head_size, num_heads, ff_dim, dropout)

    x = tf.keras.layers.GlobalAveragePooling1D(data_format="channels_first")(x)
    for dim in mlp_units:
        x = tf.keras.layers.Dense(dim, activation="relu")(x)
        x = tf.keras.layers.Dropout(mlp_dropout)(x)
    outputs = tf.keras.layers.Dense(output_shape)(x)
    return tf.keras.Model(inputs, outputs)


class Transformer(object):
    """Wrapper for Tensorflow Keras based Transformer.

    based on: https://keras.io/examples/timeseries/timeseries_transformer_classification/

    Args:

        optimizer (str): Passed to keras model.compile
        loss (str): Passed to keras model.compile
        epochs (int): Passed to keras model.fit
        batch_size (int): Passed to keras model.fit
        verbose (int): 0, 1 or 2. Passed to keras model.fit
        random_seed (int): passed to tf.random.set_seed()
    """

    def __init__(
        self,
        head_size=256,
        num_heads=4,
        ff_dim=4,
        num_transformer_blocks=4,
        mlp_units=[128],
        mlp_dropout=0.4,
        dropout=0.25,
        optimizer: str = 'adam',
        loss: str = 'huber',
        epochs: int = 50,
        batch_size: int = 32,
        verbose: int = 1,
        random_seed: int = 2020,
    ):
        self.name = 'Transformer'
        verbose = 0 if verbose < 0 else verbose
        verbose = 2 if verbose > 2 else verbose
        self.verbose = verbose
        self.random_seed = random_seed
        self.epochs = epochs
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.loss = loss

        self.head_size = head_size
        self.num_heads = num_heads
        self.ff_dim = ff_dim
        self.num_transformer_blocks = num_transformer_blocks
        self.mlp_units = mlp_units
        self.mlp_dropout = mlp_dropout
        self.dropout = dropout

    def fit(self, X, Y):
        """Train the model on dataframes of X and Y."""
        if not _has_tf:
            raise ImportError(
                "Tensorflow not available, install with pip install tensorflow."
            )
        tf.keras.backend.clear_session()
        tf.random.set_seed(self.random_seed)
        train_X = pd.DataFrame(X).to_numpy()
        train_X = train_X.reshape((train_X.shape[0], train_X.shape[1], 1))
        input_shape = train_X.shape[1:]
        if Y.ndim == 1:
            Y = Y.reshape(-1, 1)
        OUTPUT_SHAPE = Y.shape[1]

        self.model = transformer_build_model(
            input_shape,
            OUTPUT_SHAPE,
            head_size=self.head_size,
            num_heads=self.num_heads,
            ff_dim=self.ff_dim,
            num_transformer_blocks=self.num_transformer_blocks,
            mlp_units=self.mlp_units,
            mlp_dropout=self.mlp_dropout,
            dropout=self.dropout,
        )

        if self.loss == 'Huber':
            loss = tf.keras.losses.Huber()
        else:
            loss = self.loss
        optimizer = self.optimizer
        if optimizer == "adam":
            optimizer = tf.keras.optimizers.Adam(learning_rate=1e-4)
        self.model.compile(
            loss=loss,
            optimizer=optimizer,
        )

        callbacks = [
            tf.keras.callbacks.EarlyStopping(patience=10, restore_best_weights=True)
        ]

        self.model.fit(
            train_X,
            Y,
            validation_split=0.2,
            epochs=self.epochs,
            batch_size=self.batch_size,
            callbacks=callbacks,
            verbose=self.verbose,
        )
        return self

    def predict(self, X):
        """Predict on dataframe of X."""
        test = pd.DataFrame(X).to_numpy().reshape((X.shape[0], X.shape[1], 1))
        return pd.DataFrame(self.model.predict(test))



# -*- coding: utf-8 -*-
"""
Created on Fri Nov  5 13:45:01 2021

@author: Colin
"""
import timeit
import platform
from autots import model_forecast, __version__, load_linear


class Benchmark(object):
    def __init__(self):
        """Benchmark.
        Lower is better.
        Results may not be comparable across versions.
        Total runtime includes only standard package model runtimes.
        """
        self.version = __version__
        self.total_runtime = 0
        self.platform = platform.platform()
        self.node = platform.node()
        self.python_version = platform.python_version()
        self.results = {"not": "run (yet)"}
        self.avg_naive_runtime = 0
        self.sect_motif_runtime = 0
        self.nvar_runtime = 0
        self.datepart_trees_runtime = 0
        self.datepart_svm_runtime = 0
        self.theta_runtime = 0
        self.arima_runtime = 0
        self.tensorflow_rnn_runtime = 0
        self.tensorflow_cnn_runtime = 0
        self.gluonts_runtime = 0
        self.multivariate_knn_runtime = 0
        self.prophet_runtime = 0
        self.sklearn_mlp_runtime = 0

    def __repr__(self):
        """Print."""
        return f"Benchmark runtime: {self.total_runtime} see .results for details"

    def run(
        self,
        n_jobs: int = "auto",
        times: int = 3,
        random_seed: int = 123,
        base_models_only=False,
        verbose: int = 0,
    ):
        """Run benchmark.

        Args:
            n_jobs (int): passed to model_forecast for n cpus
            times (int): number of times to run benchmark models (returns avg of n times)
            random_seed (int): random seed, increases consistency
            base_models_only (bool): if True, doesn't attempt Tensorflow, GluonTS, or Prophet models
        """
        small_df = load_linear(
            long=False, shape=(200, 20), introduce_random=2, random_seed=random_seed
        )

        for _ in range(times):
            print("Beginning AverageValueNaive")
            start_time = timeit.default_timer()
            df = load_linear(
                long=False,
                shape=(200, 1000),
                introduce_random=2,
                random_seed=random_seed,
            )
            df_forecast = model_forecast(
                model_name="AverageValueNaive",
                model_param_dict={"method": "Mean"},
                model_transform_dict={
                    "fillna": "mean",
                    "transformations": {"0": "DifferencedTransformer"},
                    "transformation_params": {"0": {}},
                },
                df_train=df,
                forecast_length=12,
                frequency="D",
                prediction_interval=0.9,
                random_seed=random_seed,
                verbose=verbose,
                n_jobs=n_jobs,
            )
            self.avg_naive_runtime = (
                self.avg_naive_runtime + timeit.default_timer() - start_time
            )

            print("Beginning SectionalMotif")
            start_time = timeit.default_timer()
            df_forecast = model_forecast(
                model_name="SectionalMotif",
                model_param_dict={
                    "window": 15,
                    "point_method": "mean",
                    "distance_metric": "euclidean",
                    "include_differenced": True,
                    "k": 3,
                    "stride_size": 1,
                    "regression_type": None,
                },
                model_transform_dict={
                    "fillna": "pad",
                    "transformations": {"0": "PowerTransformer", "1": "Round"},
                    "transformation_params": {
                        "0": {},
                        "1": {"decimals": 0, "on_transform": False, "on_inverse": True},
                    },
                },
                df_train=df,
                forecast_length=12,
                frequency="D",
                prediction_interval=0.9,
                random_seed=random_seed,
                verbose=verbose,
                n_jobs=n_jobs,
            )
            self.sect_motif_runtime = (
                self.sect_motif_runtime + timeit.default_timer() - start_time
            )

            print("Beginning NVAR")
            start_time = timeit.default_timer()
            df_forecast = model_forecast(
                model_name="NVAR",
                model_param_dict={
                    "k": 2,
                    "ridge_param": 0.002,
                    "warmup_pts": 1,
                    "seed_pts": 1,
                    "seed_weighted": None,
                    "batch_size": 10,
                    "batch_method": "std_sorted",
                },
                model_transform_dict={
                    "fillna": "quadratic",
                    "transformations": {
                        "0": "Log",
                        "1": "DifferencedTransformer",
                        "2": "RollingMeanTransformer",
                        "3": "Round",
                    },
                    "transformation_params": {
                        "0": {},
                        "1": {},
                        "2": {"fixed": False, "window": 28},
                        "3": {"decimals": 0, "on_transform": False, "on_inverse": True},
                    },
                },
                df_train=df[df.columns[0:600]],
                forecast_length=12,
                frequency="D",
                prediction_interval=0.9,
                random_seed=random_seed,
                verbose=verbose,
                n_jobs=n_jobs,
            )
            self.nvar_runtime = self.nvar_runtime + timeit.default_timer() - start_time

            print("Beginning Datepart RandomForest")
            start_time = timeit.default_timer()
            df_forecast = model_forecast(
                model_name="DatepartRegression",
                model_param_dict={
                    "regression_model": {
                        "model": "RandomForest",
                        "model_params": {
                            "n_estimators": 1000,
                            "min_samples_leaf": 2,
                            "bootstrap": True,
                        },
                    },
                    "datepart_method": "simple_binarized",
                    "regression_type": None,
                },
                model_transform_dict={
                    "fillna": "ffill",
                    "transformations": {"0": "MaxAbsScaler"},
                    "transformation_params": {"0": {}},
                },
                df_train=df,
                forecast_length=15,
                frequency="D",
                prediction_interval=0.9,
                random_seed=random_seed,
                verbose=verbose,
                n_jobs=n_jobs,
            )
            self.datepart_trees_runtime = (
                self.datepart_trees_runtime + timeit.default_timer() - start_time
            )

            print("Beginning Datepart SVM")
            start_time = timeit.default_timer()
            df_forecast = model_forecast(
                model_name="DatepartRegression",
                model_param_dict={
                    "regression_model": {"model": "SVM", "model_params": {}},
                    "datepart_method": "simple",
                    "regression_type": None,
                },
                model_transform_dict={
                    "fillna": "median",
                    "transformations": {"0": "PowerTransformer"},
                    "transformation_params": {"0": {}},
                },
                df_train=df,
                forecast_length=12,
                frequency="D",
                prediction_interval=0.9,
                random_seed=random_seed,
                verbose=verbose,
                n_jobs=n_jobs,
            )
            self.datepart_svm_runtime = (
                self.datepart_svm_runtime + timeit.default_timer() - start_time
            )

            print("Beginning Theta")
            start_time = timeit.default_timer()
            df_forecast = model_forecast(
                model_name="Theta",
                model_param_dict={
                    "deseasonalize": True,
                    "difference": True,
                    "use_test": True,
                    "method": "auto",
                    "period": None,
                    "theta": 1.4,
                    "use_mle": False,
                },
                model_transform_dict={
                    "fillna": "quadratic",
                    "transformations": {
                        "0": "Log",
                        "1": "Slice",
                        "2": "RollingMeanTransformer",
                    },
                    "transformation_params": {
                        "0": {},
                        "1": {"method": 0.9},
                        "2": {"fixed": False, "window": 28},
                    },
                },
                df_train=df[df.columns[0:150]],
                forecast_length=12,
                frequency="D",
                prediction_interval=0.9,
                random_seed=random_seed,
                verbose=verbose,
                n_jobs=n_jobs,
            )
            self.theta_runtime = (
                self.theta_runtime + timeit.default_timer() - start_time
            )

            print("Beginning ARIMA")
            start_time = timeit.default_timer()
            df_forecast = model_forecast(
                model_name="ARIMA",
                model_param_dict={'p': 7, 'd': 1, 'q': 1, 'regression_type': None},
                model_transform_dict={
                    "fillna": "median",
                    "transformations": {
                        "0": "Detrend",
                    },
                    "transformation_params": {
                        "0": {'model': 'Linear'},
                    },
                },
                df_train=df[df.columns[0:60]],
                forecast_length=12,
                frequency="D",
                prediction_interval=0.9,
                random_seed=random_seed,
                verbose=verbose,
                n_jobs=n_jobs,
            )
            self.arima_runtime = (
                self.arima_runtime + timeit.default_timer() - start_time
            )

            print("Beginning Multivariate KNN")
            start_time = timeit.default_timer()
            df_forecast = model_forecast(
                model_name="MultivariateRegression",
                model_param_dict={
                    "regression_model": {
                        "model": "KNN",
                        "model_params": {"n_neighbors": 5, "weights": "uniform"},
                    },
                    "mean_rolling_periods": 30,
                    "macd_periods": None,
                    "std_rolling_periods": 7,
                    "max_rolling_periods": 60,
                    "min_rolling_periods": 60,
                    "quantile90_rolling_periods": 5,
                    "quantile10_rolling_periods": None,
                    "ewm_alpha": None,
                    "ewm_var_alpha": None,
                    "additional_lag_periods": None,
                    "abs_energy": False,
                    "rolling_autocorr_periods": None,
                    "datepart_method": "simple",
                    "polynomial_degree": None,
                    "regression_type": None,
                    "window": 7,
                    "holiday": False,
                    "probabilistic": False,
                },
                model_transform_dict={
                    "fillna": "fake_date",
                    "transformations": {"0": "MaxAbsScaler"},
                    "transformation_params": {"0": {}},
                },
                df_train=df[df.columns[0:100]],
                forecast_length=12,
                frequency="D",
                prediction_interval=0.9,
                random_seed=random_seed,
                verbose=verbose,
                n_jobs=n_jobs,
            )
            self.multivariate_knn_runtime = (
                self.multivariate_knn_runtime + timeit.default_timer() - start_time
            )
            try:
                print("Beginning MLP")
                start_time = timeit.default_timer()
                df_forecast = model_forecast(
                    model_name="WindowRegression",
                    model_param_dict={
                        "regression_model": {
                            "model": "MLP",
                            "model_params": {
                                'hidden_layer_sizes': (25, 15, 25),
                                'max_iter': 200,
                                'activation': 'tanh',
                                'solver': 'adam',
                                'early_stopping': True,
                                'learning_rate_init': 0.01,
                            },
                        },
                        "window_size": 10,
                        "input_dim": "univariate",
                        "output_dim": "forecast_length",
                        "normalize_window": False,
                        "max_windows": 5000,
                        "regression_type": None,
                    },
                    model_transform_dict={
                        "fillna": "ffill_mean_biased",
                        "transformations": {
                            "0": "MaxAbsScaler",
                        },
                        "transformation_params": {"0": {}},
                    },
                    df_train=df,
                    forecast_length=12,
                    frequency="D",
                    prediction_interval=0.9,
                    random_seed=random_seed,
                    verbose=verbose,
                    n_jobs=n_jobs,
                )
                self.sklearn_mlp_runtime = (
                    self.sklearn_mlp_runtime + timeit.default_timer() - start_time
                )
            except Exception as e:
                print(f"sklearn mlp failed with: {repr(e)}")

            if not base_models_only:
                try:
                    print("Beginning KerasRNN")
                    start_time = timeit.default_timer()
                    df_forecast = model_forecast(
                        model_name="WindowRegression",
                        model_param_dict={
                            "regression_model": {
                                "model": "KerasRNN",
                                "model_params": {
                                    "kernel_initializer": "lecun_uniform",
                                    "epochs": 50,
                                    "batch_size": 8,
                                    "optimizer": "rmsprop",
                                    "loss": "Huber",
                                    "hidden_layer_sizes": [32, 64, 32],
                                    "rnn_type": "LSTM",
                                    "shape": 1,
                                },
                            },
                            "window_size": 10,
                            "input_dim": "univariate",
                            "output_dim": "forecast_length",
                            "normalize_window": False,
                            "max_windows": 5000,
                            "regression_type": None,
                        },
                        model_transform_dict={
                            "fillna": "ffill_mean_biased",
                            "transformations": {
                                "0": "MaxAbsScaler",
                                "1": "DifferencedTransformer",
                            },
                            "transformation_params": {"0": {}, "1": {}},
                        },
                        df_train=df,
                        forecast_length=12,
                        frequency="D",
                        prediction_interval=0.9,
                        random_seed=random_seed,
                        verbose=verbose,
                        n_jobs=n_jobs,
                    )
                    self.tensorflow_rnn_runtime = (
                        self.tensorflow_rnn_runtime
                        + timeit.default_timer()
                        - start_time
                    )
                except Exception as e:
                    print(f"tensorflow failed with: {repr(e)}")

                try:
                    print("Beginning KerasCNN")
                    start_time = timeit.default_timer()
                    df_forecast = model_forecast(
                        model_name="WindowRegression",
                        model_param_dict={
                            "regression_model": {
                                "model": "KerasRNN",
                                "model_params": {
                                    "kernel_initializer": "glorot_normal",
                                    "epochs": 50,
                                    "batch_size": 8,
                                    "optimizer": "adam",
                                    "loss": "mae",
                                    "hidden_layer_sizes": [32, 32, 32],
                                    "rnn_type": "CNN",
                                    "shape": 1,
                                },
                            },
                            "window_size": 10,
                            "input_dim": "univariate",
                            "output_dim": "forecast_length",
                            "normalize_window": False,
                            "max_windows": 5000,
                            "regression_type": None,
                        },
                        model_transform_dict={
                            "fillna": "median",
                            "transformations": {
                                "0": "MaxAbsScaler",
                            },
                            "transformation_params": {"0": {}},
                        },
                        df_train=df,
                        forecast_length=12,
                        frequency="D",
                        prediction_interval=0.9,
                        random_seed=random_seed,
                        verbose=verbose,
                        n_jobs=n_jobs,
                    )
                    self.tensorflow_cnn_runtime = (
                        self.tensorflow_cnn_runtime
                        + timeit.default_timer()
                        - start_time
                    )
                except Exception as e:
                    print(f"tensorflow CNN failed with: {repr(e)}")

                try:
                    print("Beginning GluonTS")
                    start_time = timeit.default_timer()
                    df_forecast = model_forecast(
                        model_name="GluonTS",
                        model_param_dict={
                            "gluon_model": "SFF",
                            "epochs": 40,
                            "learning_rate": 0.01,
                            "context_length": 10,
                            "regression_type": None,
                        },
                        model_transform_dict={
                            "fillna": "KNNImputer",
                            "transformations": {"0": "QuantileTransformer"},
                            "transformation_params": {
                                "0": {
                                    "output_distribution": "uniform",
                                    "n_quantiles": 100,
                                }
                            },
                        },
                        df_train=df,
                        forecast_length=12,
                        frequency="D",
                        prediction_interval=0.9,
                        random_seed=random_seed,
                        verbose=verbose,
                        n_jobs=n_jobs,
                    )
                    self.gluonts_runtime = (
                        self.gluonts_runtime + timeit.default_timer() - start_time
                    )
                except Exception as e:
                    print(f"gluonts failed with: {repr(e)}")

                try:
                    print("Beginning Prophet")
                    start_time = timeit.default_timer()
                    df_forecast = model_forecast(  # noqa
                        model_name="FBProphet",
                        model_param_dict={"holiday": False, "regression_type": None},
                        model_transform_dict={
                            "fillna": "KNNImputer",
                            "transformations": {"0": "QuantileTransformer"},
                            "transformation_params": {
                                "0": {
                                    "output_distribution": "uniform",
                                    "n_quantiles": 100,
                                }
                            },
                        },
                        df_train=small_df,
                        forecast_length=12,
                        frequency="D",
                        prediction_interval=0.9,
                        random_seed=random_seed,
                        verbose=verbose,
                        n_jobs=n_jobs,
                    )
                    self.prophet_runtime = (
                        self.prophet_runtime + timeit.default_timer() - start_time
                    )
                except Exception as e:
                    print(f"prophet failed with: {repr(e)}")

        self.total_runtime = (
            self.avg_naive_runtime
            + self.sect_motif_runtime
            + self.nvar_runtime
            + self.datepart_trees_runtime
            + self.datepart_svm_runtime
            + self.multivariate_knn_runtime
            + self.theta_runtime
            + self.sklearn_mlp_runtime
            + self.arima_runtime
        ) / times
        self.results = {
            "version": self.version,
            "platform": self.platform,
            "node": self.node,
            "python_version": self.python_version,
            "n_jobs": n_jobs,
            "times": times,
            "avg_naive_runtime": self.avg_naive_runtime / times,
            "sect_motif_runtime": self.sect_motif_runtime / times,
            "nvar_runtime": self.nvar_runtime / times,
            "datepart_trees_runtime": self.datepart_trees_runtime / times,
            "datepart_svm_runtime": self.datepart_svm_runtime / times,
            "multivariate_knn_runtime": self.multivariate_knn_runtime / times,
            "theta_runtime": self.theta_runtime / times,
            "arima_runtime": self.arima_runtime / times,
            "sklearn_mlp_runtime": self.sklearn_mlp_runtime / times,
            "total_runtime": self.total_runtime,
            "tensorflow_rnn_runtime": self.tensorflow_rnn_runtime / times,
            "tensorflow_cnn_runtime": self.tensorflow_cnn_runtime / times,
            "gluonts_runtime": self.gluonts_runtime / times,
            "prophet_runtime": self.prophet_runtime / times,
        }
        return self


if __name__ == "__main__":
    import json
    import sys

    try:
        n_jobs = sys.argv[1]
        if str(n_jobs).isdigit():
            n_jobs = int(n_jobs)
    except Exception:
        n_jobs = "auto"
    try:
        times = int(sys.argv[2])
    except Exception:
        times = 3
    bench = Benchmark()
    bench = bench.run(n_jobs=n_jobs, times=times)
    print(json.dumps(bench.results, indent=1))



import datetime
import numpy as np
import pandas as pd
from autots.models.base import ModelObject, PredictionObject
from autots.tools.seasonal import date_part, seasonal_int

try:
    import tensorflow as tf
    import tensorflow_probability as tfp
    from tensorflow_probability import sts
except Exception:  # except ImportError
    _has_tf = False
else:
    _has_tf = True


class TensorflowSTS(ModelObject):
    """STS from TensorflowProbability.

    Args:
        name (str): String to identify class
        frequency (str): String alias of datetime index frequency or else 'infer'
        prediction_interval (float): Confidence interval for probabilistic forecast

        regression_type (str): type of regression (None, 'User', or 'Holiday')
    """

    def __init__(
        self,
        name: str = "TensorflowSTS",
        frequency: str = 'infer',
        prediction_interval: float = 0.9,
        regression_type: str = None,
        holiday_country: str = 'US',
        random_seed: int = 2020,
        verbose: int = 0,
        trend: str = 'local',
        seasonal_periods: int = None,
        ar_order: int = None,
        fit_method: str = 'hmc',
        num_steps: int = 200,
    ):
        ModelObject.__init__(
            self,
            name,
            frequency,
            prediction_interval,
            regression_type=regression_type,
            holiday_country=holiday_country,
            random_seed=random_seed,
            verbose=verbose,
        )
        self.seasonal_periods = seasonal_periods
        self.ar_order = ar_order
        self.trend = trend
        self.fit_method = fit_method
        self.num_steps = num_steps

    def _build_model(
        self, observed_time_series, trend='local', seasonal_periods=12, ar_order=1
    ):
        if trend == 'semilocal':
            trend = sts.SemiLocalLinearTrend(
                observed_time_series=observed_time_series, name='trend'
            )
        else:
            trend = sts.LocalLinearTrend(
                observed_time_series=observed_time_series, name='trend'
            )
        mods_list = [trend]
        if str(seasonal_periods).isdigit():
            seasonal = tfp.sts.SmoothSeasonal(
                period=seasonal_periods,
                observed_time_series=observed_time_series,
                name='seasonal',
                frequency_multipliers=[1, 2, 3],
            )
            mods_list.append(seasonal)
        if str(ar_order).isdigit():
            autoregressive = sts.Autoregressive(
                order=int(ar_order),
                observed_time_series=observed_time_series,
                name='autoregressive',
            )
            mods_list.append(autoregressive)
        model = sts.Sum(mods_list, observed_time_series=observed_time_series)
        return model

    def fit(self, df, future_regressor=[]):
        """Train algorithm given data supplied.

        Args:
            df (pandas.DataFrame): Datetime Indexed
        """
        df = self.basic_profile(df)

        demand2d = df.transpose().values
        # models expect shape: [..., num_timesteps] or [..., num_timesteps, 1]
        demand_model = self._build_model(
            demand2d,
            trend=self.trend,
            seasonal_periods=self.seasonal_periods,
            ar_order=self.ar_order,
        )

        if self.fit_method == 'variational':
            # Build the variational surrogate posteriors `qs`.
            variational_posteriors = tfp.sts.build_factored_surrogate_posterior(
                model=demand_model
            )
            q_samples = variational_posteriors.sample(int(self.num_steps))

        else:
            self.fit_method = 'hmc'
            q_samples, kernel_results = tfp.sts.fit_with_hmc(
                demand_model,
                demand2d,
                num_results=self.num_steps,
                seed=self.random_seed,
            )
        self.q_samples = q_samples
        self.demand2d = demand2d
        self.demand_model = demand_model

        self.fit_runtime = datetime.datetime.now() - self.startTime
        return self

    def predict(
        self, forecast_length: int, future_regressor=[], just_point_forecast=False
    ):
        """Generates forecast data immediately following dates of index supplied to .fit()

        Args:
            forecast_length (int): Number of periods of data to forecast ahead
            regressor (numpy.Array): additional regressor, not used
            just_point_forecast (bool): If True, return a pandas.DataFrame of just point forecasts

        Returns:
            Either a PredictionObject of forecasts and metadata, or
            if just_point_forecast == True, a dataframe of point forecasts
        """
        predictStartTime = datetime.datetime.now()
        test_index = self.create_forecast_index(forecast_length=forecast_length)
        forecast_dist = tfp.sts.forecast(
            model=self.demand_model,
            observed_time_series=self.demand2d,
            parameter_samples=self.q_samples,
            num_steps_forecast=forecast_length,
            include_observation_noise=True,
        )

        forecast = forecast_dist.mean().numpy()[..., 0]

        forecast = pd.DataFrame(
            forecast, index=self.column_names, columns=test_index
        ).transpose()

        if just_point_forecast:
            return forecast
        else:
            prediction_interval = self.prediction_interval
            # assume follows rules of normal because those are conventional
            from scipy.stats import norm

            # adj = norm.sf(abs(prediction_interval))*2
            p_int = 1 - ((1 - prediction_interval) / 2)
            adj = norm.ppf(p_int)
            forecast_scale = forecast_dist.stddev().numpy()[..., 0]
            upper_forecast = forecast.transpose().values + (forecast_scale * adj)
            lower_forecast = forecast.transpose().values - (forecast_scale * adj)
            lower_forecast = pd.DataFrame(
                lower_forecast, index=self.column_names, columns=test_index
            ).transpose()
            upper_forecast = pd.DataFrame(
                upper_forecast, index=self.column_names, columns=test_index
            ).transpose()
            # alternatively this followed by quantile
            # forecast_samples = self.forecast_dist.sample(10)[..., 0]
            predict_runtime = datetime.datetime.now() - predictStartTime
            prediction = PredictionObject(
                model_name=self.name,
                forecast_length=forecast_length,
                forecast_index=test_index,
                forecast_columns=forecast.columns,
                lower_forecast=lower_forecast,
                forecast=forecast,
                upper_forecast=upper_forecast,
                prediction_interval=self.prediction_interval,
                predict_runtime=predict_runtime,
                fit_runtime=self.fit_runtime,
                model_parameters=self.get_params(),
            )

            return prediction

    def get_new_params(self, method: str = 'random'):
        """Return dict of new parameters for parameter tuning."""
        fit_method_choice = np.random.choice(['hmc', 'variational'], size=1).item()
        num_steps_choice = np.random.choice([50, 200], size=1).item()
        ar_order_choice = np.random.choice(
            [None, 1, 2, 7], size=1, p=[0.5, 0.3, 0.1, 0.1]
        ).item()
        seasonal_periods_choice = np.random.choice(
            [None, seasonal_int()], size=1
        ).item()
        trend_choice = np.random.choice(['local', 'semilocal'], size=1).item()

        parameter_dict = {
            'fit_method': fit_method_choice,
            'num_steps': num_steps_choice,
            'ar_order': ar_order_choice,
            'seasonal_periods': seasonal_periods_choice,
            'trend': trend_choice,
        }
        return parameter_dict

    def get_params(self):
        """Return dict of current parameters."""
        parameter_dict = {
            'fit_method': self.fit_method,
            'num_steps': self.num_steps,
            'ar_order': self.ar_order,
            'seasonal_periods': self.seasonal_periods,
            'trend': self.trend,
        }
        return parameter_dict


class TFPRegressor(object):
    """Wrapper for Tensorflow Keras based RNN.

    Args:
        rnn_type (str): Keras cell type 'GRU' or default 'LSTM'
        kernel_initializer (str): passed to first keras LSTM or GRU layer
        hidden_layer_sizes (tuple): of len 1 or 3 passed to first keras LSTM or GRU layers
        optimizer (str): Passed to keras model.compile
        loss (str): Passed to keras model.compile
        epochs (int): Passed to keras model.fit
        batch_size (int): Passed to keras model.fit
        verbose (int): 0, 1 or 2. Passed to keras model.fit
        random_seed (int): passed to tf.random.set_seed()
    """

    def __init__(
        self,
        kernel_initializer: str = 'lecun_uniform',
        optimizer: str = 'adam',
        loss: str = 'negloglike',
        epochs: int = 50,
        batch_size: int = 32,
        dist: str = 'normal',
        verbose: int = 1,
        random_seed: int = 2020,
    ):
        self.name = 'TFPRegressor'
        verbose += 1
        verbose = 0 if verbose < 0 else verbose
        verbose = 2 if verbose > 2 else verbose
        self.verbose = verbose
        self.random_seed = random_seed
        self.kernel_initializer = kernel_initializer
        self.epochs = epochs
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.loss = loss  # negloglike, Huber, mae
        self.dist = dist  # normal, poisson, negbinom

    def fit(self, X, Y):
        """Train the model on dataframes of X and Y."""
        if not _has_tf:
            raise ImportError("TensorflowProbability not installed.")
        tf.keras.backend.clear_session()
        tf.keras.backend.set_floatx('float64')
        tf.random.set_seed(self.random_seed)
        train_X = pd.DataFrame(X).values
        INPUT_SHAPE = (train_X.shape[1],)
        OUTPUT_SHAPE = Y.shape[1]
        train_Y = pd.DataFrame(Y).values
        if self.dist == 'other':
            tfd = tfp.distributions
            tfk = tf.keras
            n_batches = 10
            prior = tfd.Independent(
                tfd.Normal(loc=tf.zeros(OUTPUT_SHAPE, dtype=tf.float64), scale=1.0),
                reinterpreted_batch_ndims=1,
            )
            self.nn_model = tf.keras.Sequential(
                [
                    tfk.layers.InputLayer(input_shape=INPUT_SHAPE, name="input"),
                    tfk.layers.Dense(10, activation="relu", name="dense_1"),
                    tfk.layers.Dense(
                        tfp.layers.MultivariateNormalTriL.params_size(OUTPUT_SHAPE),
                        activation=None,
                        name="distribution_weights",
                    ),
                    tfp.layers.MultivariateNormalTriL(
                        OUTPUT_SHAPE,
                        activity_regularizer=tfp.layers.KLDivergenceRegularizer(
                            prior, weight=1 / n_batches
                        ),
                        name="output",
                    ),
                ],
                name="model",
            )
        elif self.dist == 'poisson':
            self.nn_model = tf.keras.models.Sequential()
            self.nn_model.add(tf.keras.layers.Dense(10, activation="relu"))
            self.nn_model.add(
                tf.keras.layers.Dense(
                    tfp.layers.IndependentPoisson.params_size(OUTPUT_SHAPE + 1)
                )
            )
            self.nn_model.add(
                tfp.layers.DistributionLambda(
                    lambda t: tfp.distributions.Normal(
                        loc=t[..., :1], scale=1e-3 + tf.math.softplus(0.05 * t[..., 1:])
                    )
                )
            )
        else:
            self.dist = 'normal'
            self.nn_model = tf.keras.models.Sequential()
            self.nn_model.add(tf.keras.layers.Dense(OUTPUT_SHAPE + 1))
            self.nn_model.add(
                tfp.layers.DistributionLambda(
                    lambda t: tfp.distributions.Normal(
                        loc=t[..., :1], scale=1e-3 + tf.math.softplus(0.05 * t[..., 1:])
                    )
                )
            )

        if self.loss == 'negloglike':
            self.nn_model.compile(
                optimizer=self.optimizer, loss=lambda y, model: -model.log_prob(y)
            )
        else:
            if self.loss == 'Huber':
                loss = tf.keras.losses.Huber()
            else:
                loss = self.loss
            self.nn_model.compile(optimizer=self.optimizer, loss=loss)

        self.nn_model.fit(
            x=train_X,
            y=train_Y,
            epochs=self.epochs,
            batch_size=self.batch_size,
            verbose=self.verbose,
        )
        """
        nn_model.fit(x=train_X, y=train_Y, epochs=50,
                     batch_size=32)
        """
        return self

    def predict(self, X, conf_int: float = None):
        """Predict on dataframe of X."""
        test = pd.DataFrame(X).values
        if conf_int is None:
            return pd.DataFrame(self.nn_model.predict(test))
        else:
            p_int = 1 - ((1 - conf_int) / 2)
            from scipy.stats import norm

            adj = norm.ppf(p_int)

            yhat = self.nn_model(test)
            mean = self.nn_model.predict(test)
            stddev = yhat.stddev().numpy()
            mean_plus = mean - adj * stddev
            mean_minus = mean + adj * stddev
            return mean, mean_minus, mean_plus


class TFPRegression(ModelObject):
    """Tensorflow Probability regression.

    Args:
        name (str): String to identify class
        frequency (str): String alias of datetime index frequency or else 'infer'
        prediction_interval (float): Confidence interval for probabilistic forecast
        regression_type (str): type of regression (None, 'User')
    """

    def __init__(
        self,
        name: str = "TFPRegression",
        frequency: str = 'infer',
        prediction_interval: float = 0.9,
        holiday_country: str = 'US',
        random_seed: int = 2020,
        verbose: int = 1,
        kernel_initializer: str = 'lecun_uniform',
        optimizer: str = 'adam',
        loss: str = 'negloglike',
        epochs: int = 50,
        batch_size: int = 32,
        dist: str = 'normal',
        regression_type: str = None,
    ):
        ModelObject.__init__(
            self,
            name,
            frequency,
            prediction_interval,
            regression_type=regression_type,
            holiday_country=holiday_country,
            random_seed=random_seed,
            verbose=verbose,
        )
        self.verbose = verbose
        self.random_seed = random_seed
        self.kernel_initializer = kernel_initializer
        self.epochs = int(epochs)
        self.batch_size = int(batch_size)
        self.optimizer = optimizer
        self.loss = loss  # negloglike, Huber, mae
        self.dist = dist  # normal, poisson, negbinom

    def fit(self, df, future_regressor=[]):
        """Train algorithm given data supplied

        Args:
            df (pandas.DataFrame): Datetime Indexed
        """

        df = self.basic_profile(df)
        if self.regression_type == 'User':
            if (np.array(future_regressor).shape[0]) != (df.shape[0]):
                self.regression_type = None
            else:
                self.future_regressor_train = future_regressor
        self.fit_runtime = datetime.datetime.now() - self.startTime

        X = date_part(df.index, method='expanded')
        if self.regression_type == 'User':
            # if self.future_regressor_train.ndim == 1:
            #     self.future_regressor_train = np.array(self.future_regressor_train).reshape(-1, 1)
            # X = np.concatenate((X.reshape(-1, 1), self.future_regressor_train), axis=1)
            X = pd.concat(
                [X, pd.DataFrame(self.future_regressor_train).reset_index(drop=True)],
                axis=1,
            )
        y = df
        self.model = TFPRegressor(
            verbose=self.verbose,
            random_seed=self.random_seed,
            kernel_initializer=self.kernel_initializer,
            epochs=self.epochs,
            batch_size=self.batch_size,
            optimizer=self.optimizer,
            loss=self.loss,
            dist=self.dist,
        ).fit(X.values, y)
        return self

    def predict(
        self, forecast_length: int, future_regressor=[], just_point_forecast=False
    ):
        """Generates forecast data immediately following dates of index supplied to .fit()

        Args:
            forecast_length (int): Number of periods of data to forecast ahead
            regressor (numpy.Array): additional regressor, not used
            just_point_forecast (bool): If True, return a pandas.DataFrame of just point forecasts

        Returns:
            Either a PredictionObject of forecasts and metadata, or
            if just_point_forecast == True, a dataframe of point forecasts
        """
        predictStartTime = datetime.datetime.now()
        test_index = self.create_forecast_index(forecast_length=forecast_length)

        from autots.tools.seasonal import date_part

        Xf = date_part(test_index, method='expanded')
        if self.regression_type == 'User':
            # if future_regressor.ndim == 1:
            #     future_regressor = np.array(future_regressor).reshape(-1, 1)
            # Xf = np.concatenate((Xf.reshape(-1, 1), future_regressor), axis=1)
            Xf = pd.concat(
                [Xf, pd.DataFrame(future_regressor).reset_index(drop=True)], axis=1
            )
        forecast, lower_forecast, upper_forecast = self.model.predict(
            Xf.values, conf_int=self.prediction_interval
        )
        df_forecast = pd.DataFrame(forecast)
        df_forecast.columns = self.column_names
        df_forecast.index = test_index
        if just_point_forecast:
            return df_forecast
        else:
            lower_forecast = pd.DataFrame(
                lower_forecast, index=test_index, columns=self.column_names
            )
            upper_forecast = pd.DataFrame(
                upper_forecast, index=test_index, columns=self.column_names
            )
            predict_runtime = datetime.datetime.now() - predictStartTime
            prediction = PredictionObject(
                model_name=self.name,
                forecast_length=forecast_length,
                forecast_index=df_forecast.index,
                forecast_columns=df_forecast.columns,
                lower_forecast=lower_forecast,
                forecast=df_forecast,
                upper_forecast=upper_forecast,
                prediction_interval=self.prediction_interval,
                predict_runtime=predict_runtime,
                fit_runtime=self.fit_runtime,
                model_parameters=self.get_params(),
            )
            return prediction

    def get_new_params(self, method: str = 'random'):
        """Return dict of new parameters for parameter tuning."""
        init_list = [
            'glorot_uniform',
            'lecun_uniform',
            'glorot_normal',
            'RandomUniform',
            'he_normal',
        ]
        kernel_initializer = np.random.choice(init_list, size=1).item()
        epochs = np.random.choice([15, 50, 150], p=[0.2, 0.7, 0.1], size=1).item()
        batch_size = np.random.choice(
            [8, 16, 32, 72], p=[0.2, 0.2, 0.5, 0.1], size=1
        ).item()
        optimizer = np.random.choice(
            ['adam', 'rmsprop', 'adagrad'], p=[0.4, 0.5, 0.1], size=1
        ).item()
        loss = np.random.choice(
            ['negloglike', 'mae', 'Huber', 'poisson', 'mse', 'mape'],
            p=[0.3, 0.3, 0.1, 0.1, 0.1, 0.1],
            size=1,
        ).item()
        dist_choice = np.random.choice(
            a=['normal', 'poisson', 'other'], size=1, p=[0.3, 0.3, 0.4]
        ).item()
        regression_type_choice = np.random.choice(
            a=[None, 'User'], size=1, p=[0.8, 0.2]
        ).item()
        return {
            'kernel_initializer': kernel_initializer,
            'epochs': epochs,
            'batch_size': batch_size,
            'optimizer': optimizer,
            'loss': loss,  # negloglike, Huber, mae
            'dist': dist_choice,
            'regression_type': regression_type_choice,
        }

    def get_params(self):
        """Return dict of current parameters."""
        return {
            'kernel_initializer': self.kernel_initializer,
            'epochs': self.epochs,
            'batch_size': self.batch_size,
            'optimizer': self.optimizer,
            'loss': self.loss,
            'dist': self.dist,
            'regression_type': self.regression_type,
        }


"""
temp = TFPRegressor().fit(X, Y)
test = TFPRegression().fit(df)
test.predict(14).forecast

test = TensorflowSTS().fit(df).predict(14)
test.forecast
"""
